import base64
import datetime
import json
import logging
import re
import time
import traceback
import urllib.parse
from typing import Any, Callable, Dict, List, Optional, Union

import httpx
from fake_useragent import UserAgent  # type: ignore
from tenacity import retry, stop_after_attempt, wait_fixed
from ua_parser import user_agent_parser  # type: ignore

from sxm.models import LIVE_PRIMARY_HLS, XMChannel, XMLiveChannel

__all__ = [
    "HLS_AES_KEY",
    "SXMClient",
    "AuthenticationError",
    "SegmentRetrievalException",
]


SXM_APP_VERSION = "5.36.514"
SXM_DEVICE_MODEL = "EverestWebClient"
HLS_AES_KEY = base64.b64decode("0Nsco7MAgxowGvkUT8aYag==")
FALLBACK_UA = (
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:89.0) Gecko/20100101 Firefox/89.0"
)
REST_V2_FORMAT = "https://player.siriusxm.com/rest/v2/experience/modules/{}"
REST_V4_FORMAT = "https://player.siriusxm.com/rest/v4/experience/modules/{}"
SESSION_MAX_LIFE = 14400

ENABLE_NEW_CHANNELS = False


class AuthenticationError(Exception):
    """SXM Authentication failed, renew session"""

    pass


class SegmentRetrievalException(Exception):
    """failed to get HLS segment, renew session"""

    pass


class SXMClient:
    """Class to interface with SXM api and access HLS
    live streams of audio

    Parameters
    ----------
    username : :class:`str`
        SXM username
    password : :class:`str`
        SXM password
    region : :class:`str` ("US" or "CA")
        Sets your SXM account region
    user_agent : Optional[:class:`str`]
        User Agent string to use for making requests to SXM. If `None` is
        passed, it will attempt to generate one based on real browser usage
        data. Defaults to `None`.
    update_handler : Optional[Callable[[:class:`dict`], `None`]]
        Callback to be called whenever a playlist updates and new
        Live Channel data is retrieved. Defaults to `None`.

    Attributes
    ----------
    is_logged_in : :class:`bool`
        Returns if account is logged into SXM's servers
    is_session_authenticated : :class:`bool`
        Returns if session is valid and ready to use
    sxmak_token : :class:`str`
        Needs documentation
    gup_id : :class:`str`
        Needs documentation
    channels : List[:class:`XMChannel`]
        Retrieves and returns a full list of all :class:`XMChannel`
        available to the logged in account
    favorite_channels : List[:class:`XMChannel`]
        Retrieves and returns a full list of all :class:`XMChannel`
        available to the logged in account that are marked
        as favorite
    """

    last_renew: Optional[float]
    password: str
    region: str
    update_handler: Optional[Callable[[dict], None]]
    update_interval: int
    username: str

    _channels: Optional[List[XMChannel]]
    _favorite_channels: Optional[List[XMChannel]]
    _playlists: Dict[str, str]
    _ua: Dict[str, Any]

    def __init__(
        self,
        username: str,
        password: str,
        region: str = "US",
        user_agent: Optional[str] = None,
        update_handler: Optional[Callable[[dict], None]] = None,
    ):

        self._log = logging.getLogger(__file__)

        if user_agent is None:
            try:
                ua = UserAgent(use_cache_server=False)
                ua.update()
                user_agent = ua.chrome
            except Exception:
                user_agent = FALLBACK_UA
        self._ua = user_agent_parser.Parse(user_agent)

        self._reset_session()

        self.username = username
        self.password = password
        self.region = region

        self._playlists = {}
        self._channels = None
        self._favorite_channels = None

        # vars to manage session cache
        self.last_renew = None
        self.update_interval = 30

        # hook function to call whenever the playlist updates
        self.update_handler = update_handler

    @property
    def is_logged_in(self) -> bool:
        return "SXMAUTHNEW" in self._session.cookies

    @property
    def is_session_authenticated(self) -> bool:
        return (
            "AWSALB" in self._session.cookies and "JSESSIONID" in self._session.cookies
        )

    @property
    def sxmak_token(self) -> Union[str, None]:
        try:
            token = self._session.cookies["SXMAKTOKEN"]
            return token.split("=", 1)[1].split(",", 1)[0]
        except (KeyError, IndexError):
            return None

    @property
    def gup_id(self) -> Union[str, None]:
        try:
            data = self._session.cookies["SXMDATA"]
            return json.loads(urllib.parse.unquote(data))["gupId"]
        except (KeyError, ValueError):
            return None

    @property
    def channels(self) -> List[XMChannel]:
        # download channel list if necessary
        if self._channels is None:
            channels = self.get_channels()

            if len(channels) == 0:
                return []

            self._channels = []
            for channel in channels:
                self._channels.append(XMChannel.from_dict(channel))

            self._channels = sorted(self._channels, key=lambda x: int(x.channel_number))

        return self._channels

    @property
    def favorite_channels(self) -> List[XMChannel]:

        if self._favorite_channels is None:
            self._favorite_channels = [c for c in self.channels if c.is_favorite]
        return self._favorite_channels

    def login(self) -> bool:
        """Attempts to log into SXM with stored username/password"""

        self._log.debug(f"Logging in as {self.username}...")
        postdata = self._get_device_info()
        postdata.update(
            {
                "standardAuth": {
                    "username": self.username,
                    "password": self.password,
                }
            }
        )

        data = self._post("modify/authentication", postdata, authenticate=False)
        if not data:
            return False

        try:
            return data["status"] == 1 and self.is_logged_in
        except KeyError:
            self._log.error("Error decoding json response for login")
            return False

    @retry(wait=wait_fixed(3), stop=stop_after_attempt(10))
    def authenticate(self) -> bool:
        """Attempts to create a valid session for use with the client

        Raises
        ------
        AuthenticationError
            If login failed and session now needs to be reset
        """

        if not self.is_logged_in and not self.login():
            self._log.error("Unable to authenticate because login failed")
            self._reset_session()
            raise AuthenticationError("Reset session")

        data = self._post(
            "resume?OAtrial=false", self._get_device_info(), authenticate=False
        )
        if not data:
            return False

        try:
            return data["status"] == 1 and self.is_session_authenticated
        except KeyError:
            self._log.error("Error parsing json response for authentication")
            self._log.error(traceback.format_exc())
            return False

    @retry(stop=stop_after_attempt(25), wait=wait_fixed(1))
    def get_playlist(self, channel_id: str, use_cache: bool = True) -> Union[str, None]:
        """Gets playlist of HLS stream URLs for given channel ID

        Parameters
        ----------
        channel_id : :class:`str`
            ID of SXM channel to retrieve playlist for
        use_cache : :class:`bool`
            Use cached playlists for force new retrival. Defaults to `True`
        """

        url = self._get_playlist_url(channel_id, use_cache)
        if url is None:
            return None

        response = None
        try:
            response = self._make_request("GET", url, self._token_params())

            if response.status_code == 403:
                self._log.info("Received status code 403 on playlist, renewing session")
                return self.get_playlist(channel_id, False)

            if response.is_error:
                self._log.warn(
                    f"Received status code {response.status_code} on "
                    f"playlist variant"
                )
                response = None

        except httpx.RequestError as e:
            self._log.error(f"Error getting playlist: {e}")

        if response is None:
            return None

        # add base path to segments
        playlist_entries = []
        aac_path = re.findall("AAC_Data.*", url)[0]
        for line in response.text.split("\n"):
            line = line.strip()
            if line.endswith(".aac"):
                playlist_entries.append(re.sub(r"[^\/]\w+\.m3u8", line, aac_path))
            else:
                playlist_entries.append(line)

        return "\n".join(playlist_entries)

    @retry(wait=wait_fixed(1), stop=stop_after_attempt(5))
    def get_segment(self, path: str, max_attempts: int = 5) -> Union[bytes, None]:
        """Gets raw HLS segment for given path

        Parameters
        ----------
        path : :class:`str`
            SXM path
        max_attempts : :class:`int`
            Number of times to try to get segment. Defaults to 5.

        Raises
        ------
        SegmentRetrievalException
            If segments are starting to come back forbidden and session
            needs reset
        """

        url = f"{LIVE_PRIMARY_HLS}/{path}"
        res = self._session.get(url, params=self._token_params())

        if res.status_code == 403:
            raise SegmentRetrievalException(
                "Received status code 403 on segment, renew session"
            )

        if res.is_error:
            self._log.warn(f"Received status code {res.status_code} on segment")
            return None

        return res.content

    def get_channels(self) -> List[dict]:
        """Gets raw list of channel dictionaries from SXM. Each channel
        dict can be pass into the constructor of :class:`XMChannel` to turn it
        into an object"""

        channels: List[Dict[str, str]] = []

        postdata = {
            "consumeRequests": [],
            "resultTemplate": "responsive",
            "alerts": [],
            "profileInfos": [],
        }

        if ENABLE_NEW_CHANNELS:
            data = self._post(
                "get?type=2",
                postdata,
                channel_list=True,
                url_format=REST_V4_FORMAT,
            )
        else:
            data = self._post("get", postdata, channel_list=True)

        if not data:
            self._log.warn("Unable to get channel list")
            return channels

        try:
            channels = data["moduleList"]["modules"][0]["moduleResponse"][
                "contentData"
            ]["channelListing"]["channels"]
        except (KeyError, IndexError):
            self._log.error("Error parsing json response for channels")
            self._log.error(traceback.format_exc())
            return []
        return channels

    def get_channel(self, name: str) -> Union[XMChannel, None]:
        """Retrieves a specific channel from `self.channels`

        Parameters
        ----------
        name : :class:`str`
            name, id, or channel number of SXM channel to get
        """

        name = name.lower()
        for x in self.channels:
            if (
                x.name.lower() == name
                or x.id.lower() == name
                or x.channel_number == name
            ):
                return x
        return None

    def get_now_playing(self, channel: XMChannel) -> Union[Dict[str, Any], None]:
        """Gets raw dictionary of response data for the live channel.

        `data['messages'][0]['code']`
            will have the status response code from SXM

        `data['moduleList']['modules'][0]['moduleResponse']['liveChannelData']`
            will have the raw data that can be passed into
            :class:`XMLiveChannel` constructor to create an object

        Parameters
        ----------
        channel : :class:`XMChannel`
            SXM channel to look up live channel data for
        """

        params = {
            "assetGUID": channel.guid,
            "ccRequestType": "AUDIO_VIDEO",
            "channelId": channel.id,
            "hls_output_mode": "custom",
            "marker_mode": "all_separate_cue_points",
            "result-template": "web",
            "time": str(int(round(time.time() * 1000.0))),
            "timestamp": datetime.datetime.utcnow().isoformat("T") + "Z",
        }

        return self._get("tune/now-playing-live", params)

    def _reset_session(self) -> None:
        """Resets session used by client"""

        self._session_start = time.time()
        self._session = httpx.Client()
        self._session.headers.update({"User-Agent": self._ua["string"]})

    def _token_params(self) -> Dict[str, Union[str, None]]:
        return {
            "token": self.sxmak_token,
            "consumer": "k2",
            "gupId": self.gup_id,
        }

    def _get_device_info(self) -> dict:
        """Generates a dict of device info to pass to SXM"""

        browser_version = self._ua["user_agent"]["major"]
        if self._ua["user_agent"]["minor"] is not None:
            browser_version = f'{browser_version}.{self._ua["user_agent"]["minor"]}'
        if self._ua["user_agent"]["patch"] is not None:
            browser_version = f'{browser_version}.{self._ua["user_agent"]["patch"]}'

        return {
            "resultTemplate": "web",
            "deviceInfo": {
                "osVersion": self._ua["os"]["family"],
                "platform": "Web",
                "sxmAppVersion": SXM_APP_VERSION,
                "browser": self._ua["user_agent"]["family"],
                "browserVersion": browser_version,
                "appRegion": self.region,
                "deviceModel": SXM_DEVICE_MODEL,
                "clientDeviceId": "null",
                "player": "html5",
                "clientDeviceType": "web",
            },
        }

    def _make_request(
        self,
        method: str,
        path: str,
        params: Dict[str, Any],
        url_format: str = REST_V2_FORMAT,
    ) -> httpx.Response:
        if path.startswith("http"):
            url = path
        else:
            url = url_format.format(path)

        try:
            if method == "GET":
                response = self._session.get(url, params=params)
            elif method == "POST":
                response = self._session.post(url, json=params)
            else:
                raise httpx.RequestError("only GET and POST")
        except httpx.RequestError as e:
            self._log.error(
                f"An Exception occurred when trying to perform "
                f"the {method} request!"
            )
            self._log.error(f"Params: {params}")
            self._log.error(f"Method: {method}")
            self._log.error(f"Request: {e.request}")

            if isinstance(e, httpx.HTTPStatusError):
                self._log.error(f"Response: {e.response}")  # pylint: disable=no-member
            raise (e)

        return response

    def _request(
        self,
        method: str,
        path: str,
        params: Dict[str, str],
        authenticate: bool = True,
        url_format: str = REST_V2_FORMAT,
    ) -> Union[Dict[str, Any], None]:
        """Makes a GET or POST request to SXM servers"""

        method = method.upper()

        if authenticate:
            now = time.time()
            if (now - self._session_start) > SESSION_MAX_LIFE:
                self._log.info("Session exceed max time, reseting")
                self._reset_session()

            if not self.is_session_authenticated and not self.authenticate():

                self._log.error("Unable to authenticate")
                return None

        response = self._make_request(method, path, params, url_format=url_format)

        if response.is_error:
            self._log.warn(
                f"Received status code {response.status_code} for " f"path '{path}'"
            )
            self._log.warn(f"Response: {response.text}")
            return None

        try:
            return response.json()["ModuleListResponse"]
        except (KeyError, ValueError):
            self._log.error(f"Error decoding json for path '{path}'")
            return None

    def _get(
        self,
        path: str,
        params: Dict[str, str],
        authenticate: bool = True,
        url_format: str = REST_V2_FORMAT,
    ) -> Union[Dict[str, Any], None]:
        """Makes a GET request to SXM servers"""

        return self._request("GET", path, params, authenticate, url_format=url_format)

    def _post(
        self,
        path: str,
        postdata: dict,
        channel_list: bool = False,
        authenticate: bool = True,
        url_format: str = REST_V2_FORMAT,
    ) -> Union[Dict[str, Any], None]:
        """Makes a POST request to SXM servers"""
        postdata = {"moduleList": {"modules": [{"moduleRequest": postdata}]}}

        if channel_list:
            postdata["moduleList"]["modules"][0].update(
                {
                    "moduleArea": "Discovery",
                    "moduleType": "ChannelListing",
                    "moduleRequest": {"resultTemplate": "responsive"},
                }
            )

        return self._request(
            "POST", path, postdata, authenticate, url_format=url_format
        )

    def _get_playlist_url(
        self, channel_id: str, use_cache: bool = True, max_attempts: int = 5
    ) -> Union[str, None]:
        """Returns HLS live stream URL for a given `XMChannel`"""

        channel = self.get_channel(channel_id)
        if channel is None:
            self._log.info(f"No channel for {channel_id}")
            return None

        now = time.time()

        if use_cache and channel.id in self._playlists:
            if (
                self.last_renew is None
                or (now - self.last_renew) > self.update_interval
            ):

                del self._playlists[channel.id]
            else:
                return self._playlists[channel.id]

        data = self.get_now_playing(channel)
        if data is None:
            return None

        # parse response
        try:
            message = data["messages"][0]["message"]
            message_code = data["messages"][0]["code"]

        except (KeyError, IndexError):
            self._log.error("Error parsing json response for playlist")
            self._log.error(traceback.format_exc())
            return None

        # login if session expired
        if message_code == 201 or message_code == 208:
            if max_attempts > 0:
                self._log.info("Session expired, logging in and authenticating")
                if self.authenticate():
                    self._log.info("Successfully authenticated")
                    return self._get_playlist_url(
                        channel.id, use_cache, max_attempts - 1
                    )
                else:
                    self._log.error("Failed to authenticate")
                    return None
            else:
                self._log.warn("Reached max attempts for playlist")
                return None
        elif message_code == 204:
            self._log.warn("Multiple login error received, reseting session")
            self._reset_session()
            if self.authenticate():
                self._log.info("Successfully authenticated")
                return self._get_playlist_url(channel.id, use_cache, max_attempts - 1)
            else:
                self._log.error("Failed to authenticate")
                return None
        elif message_code != 100:
            self._log.warn(f"Received error {message_code} {message}")
            return None

        live_channel_raw = data["moduleList"]["modules"][0]["moduleResponse"][
            "liveChannelData"
        ]
        live_channel = XMLiveChannel.from_dict(live_channel_raw)

        self.update_interval = int(data["moduleList"]["modules"][0]["updateFrequency"])

        # get m3u8 url
        for playlist_info in live_channel.hls_infos:
            if playlist_info.size == "LARGE":
                playlist = self._get_playlist_variant_url(playlist_info.url)

                if playlist is not None:
                    self._playlists[channel.id] = playlist
                    self.last_renew = time.time()

                    if self.update_handler is not None:
                        self.update_handler(live_channel_raw)
                    return self._playlists[channel.id]
        return None

    def _get_playlist_variant_url(self, url: str) -> Union[str, None]:
        res = self._session.get(url, params=self._token_params())

        if res.is_error:
            self._log.warn(
                f"Received status code {res.status_code} on playlist "
                f"variant retrieval"
            )
            return None

        for x in res.text.split("\n"):
            if x.rstrip().endswith(".m3u8"):
                # first variant should be 256k one
                return "{}/{}".format(url.rsplit("/", 1)[0], x.rstrip())

        return None



"""HTTP Server module for sxm"""
import json
import logging
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import Type

from .client import HLS_AES_KEY, SegmentRetrievalException, SXMClient

__all__ = ["make_http_handler", "run_http_server"]


def make_http_handler(
    sxm: SXMClient, logger: logging.Logger, request_level: int = logging.INFO
) -> Type[BaseHTTPRequestHandler]:
    """
    Creates and returns a configured
    :class:`http.server.BaseHTTPRequestHandler` ready to be used
    by a :class:`http.server.HTTPServer` instance with your
    :class:`SXMClient`.

    Really useful if you want to create your own HTTP server as part
    of another application.

    Parameters
    ----------
    sxm : :class:`SXMClient`
        SXM client to use
    """

    class SXMHandler(BaseHTTPRequestHandler):
        """SXM Response handler"""

        def log_error(self, format, *args):  # noqa: A002
            logger.warn(format % args)

        def log_message(self, format, *args):  # noqa: A002
            logger.log(request_level, format % args)

        def do_GET(self):
            """Main proxy endpoint"""

            if self.path.endswith(".m3u8"):
                data = sxm.get_playlist(self.path.rsplit("/", 1)[1][:-5])
                if data:
                    self.send_response(200)
                    self.send_header("Content-Type", "application/x-mpegURL")
                    self.end_headers()
                    self.wfile.write(bytes(data, "utf-8"))
                else:
                    self.send_response(503)
                    self.end_headers()
            elif self.path.endswith(".aac"):
                segment_path = self.path[1:]
                try:
                    data = sxm.get_segment(segment_path)
                except SegmentRetrievalException:
                    sxm.reset_session()
                    sxm.authenticate()
                    data = sxm.get_segment(segment_path)

                if data:
                    self.send_response(200)
                    self.send_header("Content-Type", "audio/x-aac")
                    self.end_headers()
                    self.wfile.write(data)
                else:
                    self.send_response(503)
                    self.end_headers()
            elif self.path.endswith("/key/1"):
                self.send_response(200)
                self.send_header("Content-Type", "text/plain")
                self.end_headers()
                self.wfile.write(HLS_AES_KEY)
            elif self.path.endswith("/channels/"):
                try:
                    raw_channels = sxm.get_channels()
                except Exception:
                    raw_channels = []

                if len(raw_channels) > 0:
                    self.send_response(200)
                    self.send_header("Content-Type", "application/json; charset=utf-8")
                    self.end_headers()
                    self.wfile.write(json.dumps(raw_channels).encode("utf-8"))
                else:
                    self.send_response(403)
                    self.end_headers()
            else:
                self.send_response(404)
                self.end_headers()

    return SXMHandler


def run_http_server(
    sxm: SXMClient,
    port: int,
    ip="0.0.0.0",  # nosec
    logger: logging.Logger = None,
) -> None:
    """
    Creates and runs an instance of :class:`http.server.HTTPServer` to proxy
    SXM requests without authentication.

    You still need a valid SXM account with streaming rights,
    via the :class:`SXMClient`.

    Parameters
    ----------
    port : :class:`int`
        Port number to bind SXM Proxy server on
    ip : :class:`str`
        IP address to bind SXM Proxy server on
    """

    if logger is None:
        logger = logging.getLogger(__file__)

    httpd = HTTPServer((ip, port), make_http_handler(sxm, logger))
    try:
        logger.info(f"running SXM proxy server on http://{ip}:{port}")
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass
    httpd.server_close()



from __future__ import annotations

import datetime
import time
from typing import List, Optional, Tuple, Union

from pydantic import BaseModel  # pylint: disable=no-name-in-module

__all__ = [
    "XMArt",
    "XMImage",
    "XMCategory",
    "XMMarker",
    "XMShow",
    "XMEpisode",
    "XMEpisodeMarker",
    "XMArtist",
    "XMAlbum",
    "XMCut",
    "XMSong",
    "XMCutMarker",
    "XMPosition",
    "XMHLSInfo",
    "XMChannel",
    "XMLiveChannel",
]


LIVE_PRIMARY_HLS = "https://siriusxm-priprodlive.akamaized.net"


class XMArt(BaseModel):
    name: Optional[str]
    url: str
    art_type: str

    @staticmethod
    def from_dict(data: dict) -> XMArt:
        return XMArt(
            name=data.get("name", None),
            url=data["url"],
            art_type=data["type"],
        )


class XMImage(XMArt):
    platform: Optional[str] = None
    height: Optional[int] = None
    width: Optional[int] = None
    size: Optional[str] = None

    @staticmethod
    def from_dict(data: dict) -> XMImage:
        return XMImage(
            name=data.get("name", None),
            url=data["url"],
            art_type="IMAGE",
            platform=data.get("platform", None),
            height=data.get("height", None),
            width=data.get("width", None),
            size=data.get("size", None),
        )


class XMCategory(BaseModel):
    guid: str
    name: str
    key: Optional[str] = None
    order: Optional[int] = None
    short_name: Optional[str] = None

    @staticmethod
    def from_dict(data: dict) -> XMCategory:
        return XMCategory(
            guid=data["categoryGuid"],
            name=data["name"],
            key=data.get("key"),
            order=data.get("order"),
            short_name=data.get("shortName"),
        )


class XMMarker(BaseModel):
    guid: str
    time: int
    duration: int

    @staticmethod
    def from_dict(data: dict) -> XMMarker:
        return XMMarker(
            guid=data["assetGUID"],
            time=data["time"],
            duration=data["duration"],
        )


class XMShow(BaseModel):
    guid: str
    medium_title: str
    long_title: str
    short_description: str
    long_description: str
    arts: List[XMArt]
    # ... plus many unused

    @staticmethod
    def from_dict(data: dict) -> XMShow:
        arts: List[XMArt] = []
        for art in data.get("creativeArts", []):
            if art["type"] == "IMAGE":
                arts.append(XMImage.from_dict(art))

        return XMShow(
            guid=data["showGUID"],
            medium_title=data["mediumTitle"],
            long_title=data["longTitle"],
            short_description=data["shortDescription"],
            long_description=data["longDescription"],
            arts=arts,
        )


class XMEpisode(BaseModel):
    guid: str
    medium_title: str
    long_title: str
    short_description: str
    long_description: str
    show: XMShow
    # ... plus many unused

    @staticmethod
    def from_dict(data: dict) -> XMEpisode:
        return XMEpisode(
            guid=data["episodeGUID"],
            medium_title=data["mediumTitle"],
            long_title=data["longTitle"],
            short_description=data["shortDescription"],
            long_description=data["longDescription"],
            show=XMShow.from_dict(data["show"]),
        )


class XMEpisodeMarker(XMMarker):
    episode: XMEpisode

    @staticmethod
    def from_dict(data: dict) -> XMEpisodeMarker:
        return XMEpisodeMarker(
            guid=data["assetGUID"],
            time=data["time"],
            duration=data["duration"],
            episode=XMEpisode.from_dict(data["episode"]),
        )


class XMArtist(BaseModel):
    name: str

    @staticmethod
    def from_dict(data: dict) -> XMArtist:
        return XMArtist(name=data["name"])


class XMAlbum(BaseModel):
    title: Optional[str] = None
    arts: List[XMArt]

    @staticmethod
    def from_dict(data: dict) -> XMAlbum:
        arts: List[XMArt] = []
        for art in data.get("creativeArts", []):
            if art["type"] == "IMAGE":
                arts.append(XMImage.from_dict(art))

        return XMAlbum(title=data.get("title", None), arts=arts)


class XMCut(BaseModel):
    title: str
    artists: List[XMArtist]
    cut_type: Optional[str] = None

    @staticmethod
    def from_dict(data: dict) -> XMCut:
        artists: List[XMArtist] = []
        for artist in data["artists"]:
            artists.append(XMArtist.from_dict(artist))

        return XMCut(
            title=data["title"],
            cut_type=data.get("cutContentType", None),
            artists=artists,
        )


class XMSong(XMCut):
    album: Optional[XMAlbum] = None
    itunes_id: Optional[str] = None

    @staticmethod
    def from_dict(data: dict) -> XMSong:
        album: Optional[XMAlbum] = None
        itunes_id: Optional[str] = None

        if "album" in data:
            album = XMAlbum.from_dict(data["album"])

        for external_id in data.get("externalIds", []):
            if external_id["id"] == "iTunes":
                itunes_id = external_id["value"]

        artists: List[XMArtist] = []
        for artist in data["artists"]:
            artists.append(XMArtist.from_dict(artist))

        return XMSong(
            title=data["title"],
            cut_type=data.get("cutContentType", None),
            artists=artists,
            album=album,
            itunes_id=itunes_id,
        )


class XMCutMarker(XMMarker):
    cut: XMCut

    @staticmethod
    def from_dict(data: dict) -> XMCutMarker:
        if data["cut"].get("cutContentType", None) == "Song":
            cut: XMCut = XMSong.from_dict(data["cut"])
        else:
            cut = XMCut.from_dict(data["cut"])
        # other cuts, not done: Exp, Link., maybe more?

        return XMCutMarker(
            guid=data["assetGUID"],
            time=data["time"],
            duration=data["duration"],
            cut=cut,
        )


class XMPosition(BaseModel):
    timestamp: datetime.datetime
    position: str

    @staticmethod
    def from_dict(data: dict) -> XMPosition:
        dt_string = data["timestamp"].replace("+0000", "")
        dt = datetime.datetime.fromisoformat(dt_string)

        return XMPosition(
            timestamp=dt.replace(tzinfo=datetime.timezone.utc),
            position=data["position"],
        )


class XMHLSInfo(BaseModel):
    name: str
    url: str
    size: str
    position: Optional[XMPosition] = None
    # + unused chunks

    @staticmethod
    def from_dict(data: dict) -> XMHLSInfo:
        position: Optional[XMPosition] = None
        if "position" in data:
            position = XMPosition.from_dict(data["position"])

        return XMHLSInfo(
            name=data["name"],
            url=data["url"].replace("%Live_Primary_HLS%", LIVE_PRIMARY_HLS),
            size=data["size"],
            position=position,
        )


class XMChannel(BaseModel):
    """See `tests/sample_data/xm_channel.json` for sample"""

    guid: str
    id: str  # noqa A003
    name: str
    streaming_name: str
    sort_order: int
    short_description: str
    medium_description: str
    url: str
    is_available: bool
    is_favorite: bool
    is_mature: bool
    channel_number: int  # actually siriusChannelNumber
    images: List[XMImage]
    categories: List[XMCategory]
    # ... plus many unused

    @staticmethod
    def from_dict(data: dict):
        images: List[XMImage] = []
        for image in data["images"]["images"]:
            images.append(XMImage.from_dict(image))

        categories: List[XMCategory] = []
        for category in data["categories"]["categories"]:
            categories.append(XMCategory.from_dict(category))

        return XMChannel(
            guid=data["channelGuid"],
            id=data["channelId"],
            name=data["name"],
            streaming_name=data["streamingName"],
            sort_order=data["sortOrder"],
            short_description=data["shortDescription"],
            medium_description=data["mediumDescription"],
            url=data["url"],
            is_available=data["isAvailable"],
            is_favorite=data["isFavorite"],
            is_mature=data["isMature"],
            channel_number=data["siriusChannelNumber"],
            images=images,
            categories=categories,
        )

    @property
    def pretty_name(self) -> str:
        """Returns a formated version of channel number + channel name"""
        return f"#{self.channel_number} {self.name}"


class XMLiveChannel(BaseModel):
    """See `tests/sample_data/xm_live_channel.json` for sample"""

    id: str  # noqa A003
    hls_infos: List[XMHLSInfo]
    custom_hls_infos: List[XMHLSInfo]
    episode_markers: List[XMEpisodeMarker]
    cut_markers: List[XMCutMarker]
    _song_cuts: Optional[List[XMCutMarker]] = None
    tune_time: Optional[int] = None
    # ... plus many unused

    @staticmethod
    def from_dict(data: dict) -> XMLiveChannel:
        hls_infos: List[XMHLSInfo] = []
        for info in data["hlsAudioInfos"]:
            hls_infos.append(XMHLSInfo.from_dict(info))

        custom_hls_infos, tune_time = XMLiveChannel._get_custom_hls_infos(
            data["customAudioInfos"]
        )
        episode_markers, cut_markers = XMLiveChannel._get_markers(data["markerLists"])

        return XMLiveChannel(
            id=data["channelId"],
            hls_infos=hls_infos,
            custom_hls_infos=custom_hls_infos,
            tune_time=tune_time,
            episode_markers=episode_markers,
            cut_markers=cut_markers,
        )

    @staticmethod
    def _get_custom_hls_infos(custom_infos) -> Tuple[List[XMHLSInfo], Optional[int]]:
        custom_hls_infos: List[XMHLSInfo] = []
        tune_time: Optional[int] = None

        for info in custom_infos:
            custom_hls_infos.append(XMHLSInfo.from_dict(info))

        for hls_info in custom_hls_infos:
            if (
                hls_info.position is not None
                and hls_info.position.position == "TUNE_START"
            ):

                timestamp = hls_info.position.timestamp.timestamp()
                tune_time = int(timestamp * 1000)

        return custom_hls_infos, tune_time

    @staticmethod
    def _get_markers(marker_lists) -> Tuple[List[XMEpisodeMarker], List[XMCutMarker]]:
        episode_markers: List[XMEpisodeMarker] = []
        cut_markers: List[XMCutMarker] = []

        for marker_list in marker_lists:
            # not including future-episodes as they are missing metadata
            if marker_list["layer"] == "episode":
                episode_markers = XMLiveChannel._get_episodes(marker_list["markers"])
            elif marker_list["layer"] == "cut":
                cut_markers = XMLiveChannel._get_cuts(marker_list["markers"])

        return episode_markers, cut_markers

    @staticmethod
    def _get_episodes(markers) -> List[XMEpisodeMarker]:
        episode_markers: List[XMEpisodeMarker] = []

        for marker in markers:
            episode_markers.append(XMEpisodeMarker.from_dict(marker))

        episode_markers = XMLiveChannel.sort_markers(episode_markers)  # type: ignore
        return episode_markers

    @staticmethod
    def _get_cuts(markers) -> List[XMCutMarker]:
        cut_markers: List[XMCutMarker] = []

        for marker in markers:
            if "cut" in marker:
                cut_markers.append(XMCutMarker.from_dict(marker))

        cut_markers = XMLiveChannel.sort_markers(cut_markers)  # type: ignore
        return cut_markers

    @property
    def song_cuts(self) -> List[XMCutMarker]:
        """Returns a list of all `XMCut` objects that are for songs"""

        if self._song_cuts is None:
            self._song_cuts = []
            for cut in self.cut_markers:
                if isinstance(cut.cut, XMSong):
                    self._song_cuts.append(cut)

        return self._song_cuts

    @staticmethod
    def sort_markers(markers: List[XMMarker]) -> List[XMMarker]:
        """Sorts a list of `XMMarker` objects"""
        return sorted(markers, key=lambda x: x.time)

    def _latest_marker(
        self, marker_attr: str, now: Optional[int] = None
    ) -> Union[XMMarker, None]:
        """Returns the latest `XMMarker` based on type relative to now"""

        markers = getattr(self, marker_attr)
        if markers is None:
            return None

        if now is None:
            now = int(time.time() * 1000)

        latest = None
        for marker in markers:
            if now > marker.time:
                latest = marker
            else:
                break
        return latest

    def get_latest_episode(
        self, now: Optional[int] = None
    ) -> Union[XMEpisodeMarker, None]:
        """Returns the latest :class:`XMEpisodeMarker` based
        on type relative to now

        Parameters
        ----------
        now : Optional[:class:`int`]
            Timestamp in milliseconds from Epoch to be considered `now`
        """
        return self._latest_marker("episode_markers", now)  # type: ignore

    def get_latest_cut(self, now: Optional[int] = None) -> Union[XMCutMarker, None]:
        """Returns the latest :class:`XMCutMarker` based
        on type relative to now

        Parameters
        ----------
        now : Optional[:class:`int`]
            Timestamp in milliseconds from Epoch to be considered `now`
        """
        return self._latest_marker("cut_markers", now)  # type: ignore



# -*- coding: utf-8 -*-

"""Console script for sxm."""
import logging
from enum import Enum

import typer

from sxm import SXMClient, run_http_server


class RegionChoice(str, Enum):
    US = "US"
    CA = "CA"


OPTION_USERNAME = typer.Option(
    ..., "--username", "-U", help="SXM username", prompt=True, envvar="SXM_USERNAME"
)
OPTION_PASSWORD = typer.Option(
    ...,
    "--password",
    "-P",
    help="SXM password",
    prompt=True,
    hide_input=True,
    envvar="SXM_PASSWORD",
)
OPTION_LIST_CHANNELS = typer.Option(
    False, "--list-channels", "-l", help="List all available SXM channels and exit"
)
OPTION_PORT = typer.Option(
    9999, "--port", "-p", help="Port to run SXM server on", envvar="SXM_PORT"
)
OPTION_HOST = typer.Option(
    "127.0.0.1", "--host", "-h", help="IP to bind SXM server to", envvar="SXM_HOST"
)
OPTION_VERBOSE = typer.Option(
    False, "--verbose", "-v", help="Enable debug logging", envvar="SXM_DEBUG"
)
OPTION_REGION = typer.Option(
    RegionChoice.US,
    "--region",
    "-r",
    help="Sets the SXM client's region",
    envvar="SXM_REGION",
)


def main(
    username: str = OPTION_USERNAME,
    password: str = OPTION_PASSWORD,
    do_list: bool = OPTION_LIST_CHANNELS,
    port: int = OPTION_PORT,
    host: str = OPTION_HOST,
    verbose: bool = OPTION_VERBOSE,
    region: RegionChoice = OPTION_REGION,
) -> int:
    """SXM proxy command line application."""

    if verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    sxm = SXMClient(username, password, region=region)
    if do_list:
        l1 = max(len(x.id) for x in sxm.channels)
        l2 = max(len(str(x.channel_number)) for x in sxm.channels)
        l3 = max(len(x.name) for x in sxm.channels)

        typer.echo(f"{'ID'.ljust(l1)} | {'Num'.ljust(l2)} | {'Name'.ljust(l3)}")
        for channel in sxm.channels:
            channel_id = channel.id.ljust(l1)[:l1]
            channel_num = str(channel.channel_number).ljust(l2)[:l2]
            channel_name = channel.name.ljust(l3)[:l3]
            typer.echo(f"{channel_id} | {channel_num} | {channel_name}")
    else:
        run_http_server(sxm, port, ip=host)
    return 0



# -*- coding: utf-8 -*-
"""Module entrypoint for sxm."""

import typer
from dotenv import load_dotenv

from sxm.cli import main


def start():
    """Wrapper around typer entrypoint"""
    load_dotenv()

    typer.run(main)


if __name__ == "__main__":
    start()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['sxm']

package_data = \
{'': ['*']}

install_requires = \
['fake-useragent',
 'httpx',
 'pydantic',
 'python-dotenv',
 'tenacity',
 'typer',
 'ua-parser']

extras_require = \
{'dev': ['bandit',
         'black',
         'flake8-bugbear',
         'flake8-builtins',
         'flake8-commas',
         'flake8-comprehensions',
         'flake8',
         'flit',
         'isort',
         'mypy',
         'pip-tools',
         'pylint',
         'pyproject-flake8',
         'rstcheck',
         'typing-extensions~=3.7.4.1'],
 'doc': ['furo', 'snooty-lextudio', 'sphinx'],
 'test': ['pytest',
          'pytest-cov',
          'pytest-socket',
          'pytest-sugar',
          'pytest-clarity',
          'tox',
          'coverage']}

entry_points = \
{'console_scripts': ['sxm = sxm.__main__:start']}

setup(name='sxm',
      version='0.2.3',
      description='Top-level package for sxm.',
      author='AngellusMortis',
      author_email='cbailey@mort.is',
      url='https://github.com/AngellusMortis/sxm-client',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.7',
     )



# -*- coding: utf-8 -*-

"""Top-level package for sxm."""

from sxm.client import (
    HLS_AES_KEY,
    AuthenticationError,
    SegmentRetrievalException,
    SXMClient,
)
from sxm.http import make_http_handler, run_http_server

__author__ = """AngellusMortis"""
__email__ = "cbailey@mort.is"
__version__ = "0.2.3"
__all__ = [
    "AuthenticationError",
    "HLS_AES_KEY",
    "make_http_handler",
    "run_http_server",
    "SegmentRetrievalException",
    "SXMClient",
]



from unittest.mock import patch

from sxm.models import XMLiveChannel


def test_channel(sxm_client):
    channel = sxm_client.get_channel("octane")
    assert channel.guid == "0fed9647-cc82-24d7-526d-98762e8a52cd"
    assert channel.id == "octane"
    assert channel.streaming_name == "Cutting Edge Of New Hard Rock"
    assert channel.sort_order == 410
    assert channel.short_description == "Cutting Edge Of New Hard Rock"
    assert channel.medium_description == (
        "The cutting edge of new hard rock, featuring the next generation of "
        "headbangers destined to be headliners."
    )
    assert channel.url == "https://player.siriusxm.com/live/Octane"
    assert channel.is_available is True
    assert channel.is_favorite is False
    assert channel.is_mature is False
    assert channel.channel_number == 37


def test_channels(sxm_client):
    channels = sxm_client.channels

    assert len(channels) == 363

    channel = channels[0]
    assert channel.guid == "194adbca-34d6-cb94-b153-3488ee563308"
    assert channel.id == "siriushits1"
    assert channel.streaming_name == "Today's Pop Hits"
    assert channel.sort_order == 10
    assert channel.short_description == "Today's Pop Hits"
    assert channel.medium_description == (
        "The Most Hit-Music, featuring The Morning Mash Up, Hits 1 in Hollywood, "
        "The Weekend Countdown and Hit-Bound."
    )
    assert channel.url == "https://player.siriusxm.com/live/SiriusXMHits1"
    assert channel.is_available is True
    assert channel.is_favorite is False
    assert channel.is_mature is False
    assert channel.channel_number == 2


def test_live_channel(sxm_client):
    data = sxm_client.get_now_playing("octane")

    channel = XMLiveChannel.from_dict(
        data["moduleList"]["modules"][0]["moduleResponse"]["liveChannelData"]
    )

    assert channel.id == "octane"

    with patch("sxm.models.time") as mock_time:
        mock_time.time.return_value = 1630000000
        cut = channel.get_latest_cut()

    assert cut.cut.title == "Ten Thousand Fists"
    assert cut.cut.cut_type == "Song"
    assert cut.cut.itunes_id is None
    assert cut.cut.album.title == "Ten Thousand Fists"
    assert len(cut.cut.artists) == 1
    assert cut.cut.artists[0].name == "Disturbed"
